from typing import Callable, Optional
import numpy as np

from quple.data_encoding.pauli_z_expansion import PauliZExpansion

class SecondOrderExpansion(PauliZExpansion):
    def __init__(self, feature_dimension: int,
                 depth: int=2, 
                 encoding_map:Optional[Callable[[np.ndarray], float]] = None,
                 name:str=None):
        '''Create Second Order Expansion feature map
        Args:
            feature_dimension: dimension of data to be encoded (=number of qubits in the circuit)
            depth: the number of repetition of the encoding circuit
            encoding_map: data mapping function from R^(feature_dimension) to R
            name: name of circuit
        '''        
        super().__init__(feature_dimension, depth, z_order=2,
                         encoding_map=encoding_map, name=name)


from abc import ABC, abstractmethod
import itertools
from typing import (Any, Callable, cast, Dict, FrozenSet, Iterable, Iterator,
                    List, Optional, overload, Sequence, Set, Tuple, Type,
                    TYPE_CHECKING, TypeVar, Union)
import re
import numpy as np 
from pdb import set_trace
import sympy

import cirq   
from cirq import  devices, GridQubit 
from cirq.circuits import InsertStrategy

from quple.circuit.qubit_register import QubitRegister
from quple.components.parameter_table import ParameterTable


# InsertStrategy.EARLIEST, InsertStrategy.NEW, InsertStrategy.INLINE and InsertStrategy.NEW_THEN_INLINE

kGateMapping = {
        "H": cirq.H, # Hadamard gate
        "I": cirq.I,  # one-qubit Identity gate
        "S": cirq.S, # Clifford S gate
        "T": cirq.T, # non-Clifford T gate
        'X': cirq.X, # Pauli-X gate
        "Y": cirq.Y, # Pauli-Y gate
        "Z": cirq.Z, # Pauli-Z gate
        "PauliX": cirq.X, # Pauli-X gate
        "PauliY": cirq.Y, # Pauli-Y gate
        "PauliZ": cirq.Z, # Pauli-Z gate
        "CX": cirq.CX, # Controlled-NOT gate
        "CNOT": cirq.CNOT, # Controlled-NOT gate
        "CZ": cirq.CZ, # Controlled-Z gate
        "XX": cirq.XX, # tensor product of two X gates
        "YY": cirq.YY, # tensor product of two Y gates
        "ZZ": cirq.ZZ, # tensor product of two Z gates
        "RX": cirq.rx, # rotation along X axis
        "RY": cirq.ry, # rotation along Y axis
        "RZ": cirq.rz, # rotation along Z axis
        "CCNOT": cirq.CCNOT, # Toffoli gate
        "CCX": cirq.CCX, # Toffoli gate
        "Toffoli": cirq.TOFFOLI, # Toffoli gate
        "SWAP": cirq.SWAP, # SWAP gate
        "CSWAP": cirq.CSWAP, # Controlled SWAP gate
        "ISWAP": cirq.ISWAP, # ISWAP gate
        "RISWAP": cirq.riswap, #Rotation ISWAP gate (X⊗X + Y⊗Y)
        "FSim": cirq.FSimGate, # Fermionic simulation gate
        "Fredkin": cirq.FREDKIN, # Controlled SWAP gate
    }


class QuantumCircuit(cirq.Circuit):
    def __init__(self, n_qubit:int=0, name:str=None,
                insert_strategy:InsertStrategy=None) -> None:
        super().__init__()
        self._name = 'qc' or name
        self._qubits = QubitRegister(n_qubit)
        self._parameter_table = ParameterTable()
        self._insert_strategy = insert_strategy or InsertStrategy.INLINE
        
    @property
    def name(self) -> str:
        return self._name
    
    @property
    def n_qubit(self) -> int:
        return self._qubits.size
    

    @n_qubit.setter
    def n_qubit(self, value:int):
        self._qubits = QubitRegister(value)
    
    @property
    def circuit(self):
        return self
    
    @property
    def parameter_table(self):
        return self._parameter_table
    
    @property
    def qubits(self):
        return self._qubits
    
    @property
    def insert_strategy(self):
        return self._insert_strategy

    @property
    def diagram(self):
        print(self)
    
    @staticmethod
    def _parse_gate_operation(gate_name:str):
        if gate_name in kGateMapping:
            return kGateMapping[gate_name]
        else:
            raise KeyError("Unknown gate operation {}".format(gate_name))
        

    def apply_gate_operation(self, operation:[str, cirq.Gate], qubit_expr):
        if isinstance(operation, str):
            operation = QuantumCircuit._parse_gate_operation(operation)
        qubit_sequence = self._qubits.get(qubit_expr)
        strategy = self.insert_strategy
        if all(isinstance(qubits, GridQubit) for qubits in qubit_sequence):
            self.append([operation(qubit) for qubit in qubit_sequence], strategy=strategy)
        elif all(isinstance(qubits, tuple) for qubits in qubit_sequence):
            self.append([operation(*qubits) for qubits in qubit_sequence], strategy=strategy)
        else:
            raise ValueError('Inconsistent qubit representation: {}'.format(qubit_sequence))
            
    def get_param_resolver(self, param_values: Dict):
        return cirq.ParamResolver(param_values)
    
    def _test01(self, repetitions=500):
        simulator.run(self, param_resolver, repetitions=repetitions)

    def S(self, qubit_expr):
        '''Clifford-S gate
        '''
        self.apply_gate_operation(cirq.ops.S, qubit_expr)

    def T(self, qubit_expr):
        '''Non-Clifford-T gate
        '''
        self.apply_gate_operation(cirq.ops.T, qubit_expr)


    def H(self, qubit_expr):
        '''Hadamard gate
        '''
        self.apply_gate_operation(cirq.ops.H, qubit_expr)
    
    def RX(self, theta:Union[int, float], qubit_expr):
        self.apply_gate_operation(cirq.ops.rx(theta), qubit_expr)

    def RY(self, theta:Union[int, float], qubit_expr):
        self.apply_gate_operation(cirq.ops.ry(theta), qubit_expr)

    def RZ(self, theta:Union[int, float], qubit_expr):
        self.apply_gate_operation(cirq.ops.rz(theta), qubit_expr)                
    
    def SWAP(self, qubit_expr):
        '''SWAP gate
        '''
        self.apply_gate_operation(cirq.ops.SWAP, qubit_expr)
    

    def PauliX(self, qubit_expr):
        self.apply_gate_operation(cirq.X, qubit_expr)
        
    def PauliY(self, qubit_expr):
        self.apply_gate_operation(cirq.Y, qubit_expr)

    def PauliZ(self, qubit_expr):
        self.apply_gate_operation(cirq.Z, qubit_expr)
        

    def CNOT(self, qubit_expr):
        '''Controlled NOT gate
        '''
        self.apply_gate_operation(cirq.ops.CNOT, qubit_expr)
    
    def Toffoli(self, qubit_expr):
        '''Toffoli gate
        '''
        self.apply_gate_operation(cirq.ops.TOFFOLI, qubit_expr)

    def PhaseShift(self, phi, qubit_expr):
        '''PhaseShit gate
        Perform PhaseShift gate operation
        '''
        self.apply_gate_operation(cirq.ZPowGate(exponent=phi / np.pi), qubit_expr)
            
    def CX(self, qubit_expr):
        '''Equivalent to CNOT gate
        Perform controlled-X gate operation
        '''
        self.apply_gate_operation(cirq.ops.CX, qubit_expr)
  
    def CZ(self, qubit_expr):
        '''Controlled-Z gate
        Perform controlled-X gate operation
        '''
        self.apply_gate_operation(cirq.ops.CZ, qubit_expr)
        
    def clear(self) -> None:
        '''
        clear the content of the circuit
        '''
        self._moments = []
        self._parameter_table = ParameterTable()
        
    def _entangled_qubit_pairing(self, qubits: Sequence[int], 
        *args, **kwargs) -> List[Tuple[int]]:
        '''determines how qubits are paired in the entanglement operation
        can be overridden
        Args:
            qubits: qubits to be entangled
        Return:
            list of pair of qubit indices for entanglement
        Example:
        >>> cq = EntanglementCircuit(n_qubit=5)
        >>> cq._entangled_qubit_pairing((1,3,4)) #entangle qubits 1, 3 and 4
        [(1,3), (3,4)]
        '''
        pairing_indices = [(qubits[i], qubits[i+1]) for i in range(len(qubits)-1)]
        
        return pairing_indices
    
    def entangle(self, qubits:Sequence[int],
                 inverse:bool=False,
                 gate:cirq.Gate=cirq.ops.CNOT):
        '''entangle qubits in a quantum cirquit
        Args:
            qubits: qubits to be entangled
            inverse: reverse the order of operation
            gate: gate operation that entangles the qubits
        Example:
        >>> cq = QuantumCircuit(n_qubit=5)
        >>> cq.entangle((1,2))
        >>> cq.entangle((0,3,4))
                       ┌──┐
            (0, 0): ─────@────────
                         │
            (1, 0): ────@┼────────
                        ││
            (2, 0): ────X┼────────
                         │
            (3, 0): ─────X────@───
                              │
            (4, 0): ──────────X───
                       └──┘
        '''
        
        pairing_indices = self._entangled_qubit_pairing(qubits)
        qubit_pairs = self._qubits.get(pairing_indices)
        if inverse:
            qubit_pairs = qubit_pairs[::-1]
        self.append([gate(*qpair) for qpair in qubit_pairs])  
        
    def new_layer(self, name:str=None, *args, **kwargs):
        '''New circuit layer
        Create a circuit block that is a new instance of the current circuit type
        '''
        return self.__class__(self.n_qubit, name=name, *args, **kwargs)
            


from typing import List, Union, Optional, Callable, Sequence
import numpy as np
from pdb import set_trace

import cirq

from quple.circuit.quantum_circuit import QuantumCircuit
from quple.data_encoding.encoding_circuit import EncodingCircuit
from quple.circuit.templates.pauli_block import PauliBlock

class PauliExpansion(EncodingCircuit):
    '''PauliExpansion feature map

    '''    
    def __init__(self, feature_dimension: int,
                 copies: int=2, paulis:List[str] = ['Z', 'ZZ'],
                 encoding_map:Optional[Callable[[np.ndarray], float]] = None,
                 name:str=None):
        '''Create PauliExpansion encoder circuit
        Args:
            feature_dimension: dimension of data to be encoded (=number of qubits in the circuit)
            depth: the number of repetition of the encoding circuit
            paulis: pauli operations to be performed on each entangling block
            encoding_map: data mapping function from R^(feature_dimension) to R
            name: name of circuit
        '''
        PauliExpansion._validate_paulis(paulis)
        pauli_blocks = [PauliBlock(pauli) for pauli in paulis]
        super().__init__(feature_dimension, copies, 
                         entanglement_blocks=pauli_blocks,
                         encoding_map=encoding_map, name=name)
        self.paulis = paulis
        
    @staticmethod
    def _validate_paulis(paulis:List[str]):
        for pauli_str in paulis:
            for pauli in pauli_str:
                if pauli not in ['Z','X','Y','I']:
                    raise ValueError('Invalid Pauli operation: {}'.format(pauli))


from quple.components.parameter_table import ParameterTable
from quple.circuit.qubit_register import QubitRegister
from quple.circuit.quantum_circuit import QuantumCircuit
from quple.circuit.parameterised_circuit import ParameterisedCircuit
from quple.circuit.templates.template_circuit_block import TemplateCircuitBlock



from abc import ABC, abstractmethod
from typing import Optional, List, Union, Callable
import numpy as np

import cirq

from quple import ParameterisedCircuit
from quple.data_encoding.encoding_maps import self_product

class EncodingCircuit(ParameterisedCircuit, ABC):
    '''The data encoding circuit
    A parametrised quantum circuit for encoding classical data into quantum states
    Examples:
    >>> cq = EncodingCircuit(5, depth=1, param_symbol='x')
    '''
    def __init__(self, feature_dimension:int, copies:int=1,
                 encoding_map:Optional[Callable[[np.ndarray], float]] = None,
                 parameter_symbol:Optional[str]=None, name:str=None):
        '''Create a new Pauli expansion circuit.
        Args:
            feature_dimension: dimension of data to be encoded (=number of qubits in the circuit)
            depth: the number of repetition of the encoding circuit
            encoding_map: data mapping function from R^(feature_dimension) to R
            parameter_symbol: the symbol for the unresolved parameter array
            name: name of circuit
        '''
        super().__init__(n_qubit=feature_dimension, copies=copies, 
                         parameter_symbol=parameter_symbol, name=name)
        self._feature_dimension = feature_dimension
        self._encoding_map = encoding_map or self_product
        self._parameter_table.append_array(self._parameter_symbol, feature_dimension)
        
    def get_parameters(self, indices:List[int]):
        '''Obtain parameter symbols corresponding to the array indices
        Args:
            indices: array indices of the parameter
        Return:
            parameter symbols
        Example:
        >>>  cq.get_parameters((1,2))
        (x[1], x[2])
        '''
        return self._parameter_table[self._parameter_symbol][indices]

    def encode_parameters(self, indices:List[int]):
        '''Obtain the encoded value for the given parameters using the encoding map
        Args:
            indices: indices of parameter array which corresponds to the indices of qubits in the circuit
        Returns:
            symbolic representation of the encoded data using the encoding map
        Example:
        >>> from quple.data_encoding.PauliExpansion import PauliExpansion
        >>> cq = PauliExpansion(feature_dimension=5, depth=1)
        >>> cq.encode_parameters(indices=(0,2,4))
        >>> (np.pi-x[0])*(np.pi-x[2])*(np.pi-x[4])
        '''
        parameters = self.get_parameters(indices)
        encoded_value = self.encoding_map(parameters)
        return encoded_value

    @property
    def parameter_symbol(self):
        return self._parameter_symbol
    
    @property
    def feature_dimension(self):
        return self._feature_dimension
    
    
    @property
    def encoding_map(self):
        return self._encoding_map


from typing import List, Tuple
import itertools
    
def nearest_neighbor(n:int, m:int=2) -> List[Tuple[int]]:
    return [tuple(range(i, i+m)) for i in range(n - m + 1)]    

def cyclic(n:int, m:int=2) -> List[Tuple[int]]:
    return nearest_neighbor(n,m) + [tuple(range(n - m + 1, n)) + (0,)]

def fully_connected(n:int, m:int=2) -> List[Tuple[int]]:
    return list(itertools.combinations(list(range(n)), m))

def all_to_all(n:int, m:int=2) -> List[Tuple[int]]:
    return list(itertools.permutations(list(range(n)), m))

def star(n:int, m:int=2) -> List[Tuple[int]]:
    if m != 2:
        raise ValueError('the connectiviy graphs of the "star" method'
                         ' requires 2 qubits in an interaction unit'
                        ' but {} is given'.format(m))
    return [(0, i+1) for i in range(n - m + 1)]


connectivity_graph = {
    'nearest_neighbor': nearest_neighbor,
    'linear': nearest_neighbor,
    'cyclic': cyclic,
    'circular': cyclic,
    'full': fully_connected,
    'fully_connected': fully_connected,
    'all_to_all': all_to_all,
    'star': star
}


from cirq import GridQubit

from typing import (Any, Callable, cast, Dict, FrozenSet, Iterable, Iterator,
                    List, Optional, overload, Sequence, Set, Tuple, Type,
                    TYPE_CHECKING, TypeVar, Union)    


    
class QubitRegister():
    def __init__(self, size:int=0):
        self._size = size
        self._qubits = [GridQubit(i, 0) for i in range(size)]

        
    def __getitem__(self, key: int) -> GridQubit:
        return self._qubits[key]
    
    @staticmethod
    def _is_unique_qubit_set(qubit_set: Union[List[Tuple[GridQubit]],Tuple[GridQubit]]):
        if isinstance(qubit_set, tuple):
            return len(set(qubit_set)) == len(qubit_set)
        elif isinstance(qubit_set, list):
            return all(len(set(qubit_subset)) == len(qubit_subset) for qubit_subset in qubit_set)
        return False
        

    @staticmethod
    def _parse_qubit_expression(qubit_expr, target_qubits:Sequence[GridQubit],
                               unique_set=True) -> Union[List[GridQubit], List[Tuple[GridQubit]]]:
        '''
        return list of qubits or list of tuples of qubits based on its indicess
        '''
        def check_uniqueness(qubit_sequence):
            for qubits in qubit_sequence:
                if isinstance(qubits, tuple) and \
                (not QubitRegister._is_unique_qubit_set(qubits)):
                    raise ValueError('Qubits in the set {} are not unique'.format(qubits))
                
        resolved_qubits = None
        try:
            if isinstance(qubit_expr, GridQubit):
                resolved_qubits = [qubit_expr]
            elif isinstance(qubit_expr, int):
                resolved_qubits = [target_qubits[qubit_expr]]
            elif isinstance(qubit_expr, (range,list)) and \
            all(isinstance(expr, int) for expr in qubit_expr):
                resolved_qubits = [target_qubits[i] for i in qubit_expr]
            elif isinstance(qubit_expr, list) and \
            all(isinstance(expr, tuple) for expr in qubit_expr):
                resolved_qubits = [tuple(target_qubits[i] for i in qtuples) for qtuples in qubit_expr] 
            elif isinstance(qubit_expr, list) and \
            all(isinstance(expr, GridQubit) for expr in qubit_expr):
                resolved_qubits = qubit_expr
            elif isinstance(qubit_expr, tuple):
                resolved_qubits = [tuple(target_qubits[i] for i in qubit_expr)]
                QubitRegister._is_unique_qubit_set(resolved_qubits)
            elif isinstance(qubit_expr, slice):
                resolved_qubits = target_qubits[qubit_expr]
            else:
                raise ValueError('Unsupported qubit expression {} ({})'.format(qubit_expr, type(qubit_expr)))
        except IndexError:
                raise IndexError('Qubit index out of range.') from None
        except TypeError:
                raise IndexError('Qubit index must be an integer') from None
        if unique_set:      
            check_uniqueness(resolved_qubits)
        return resolved_qubits
    
    
    def get(self, qubit_expr) -> Union[List[GridQubit], List[Tuple[GridQubit]]]:
        return QubitRegister._parse_qubit_expression(qubit_expr, self._qubits)
    
    @property
    def size(self) -> int:
        return self._size

    @property
    def qubits(self) -> List[GridQubit]:
        return self._qubits    


from abc import ABC, abstractmethod
from inspect import signature
from typing import (Any, Callable, cast, Dict, FrozenSet, Iterable, Iterator,
                    List, Optional, overload, Sequence, Set, Tuple, Type,
                    TYPE_CHECKING, TypeVar, Union)
import sympy as sp

import cirq
from cirq.ops.gate_features import SingleQubitGate, TwoQubitGate

from quple.circuit.quantum_circuit import QuantumCircuit
from quple.circuit.templates.template_circuit_block import TemplateCircuitBlock
from quple.components.parameter_table import ParameterTable
from quple.components.connectivity_graphs import connectivity_graph
from pdb import set_trace
#D. Zhu, Training of quantum circuits on a hybrid quantum computer


#Devices generally operate in the Z basis, so that rotations around the Z axis will become book-keeping measures rather than physical operations on the device.

class ParameterisedCircuit(QuantumCircuit, ABC):
    def __init__(self, n_qubit:int, copies:int=1,
                rotation_blocks:Optional[Union[str, cirq.Gate, Callable, 'TemplateCircuitBlock',
                                               List[str],List[cirq.Gate],List[Callable],
                                               List['TemplateCircuitBlock']]] =None,
                entanglement_blocks:Optional[Union[str, cirq.Gate, Callable, 'TemplateCircuitBlock',
                                               List[str],List[cirq.Gate],List[Callable],
                                               List['TemplateCircuitBlock']]] =None,
                entangle_strategy:Optional[Union[str,List[str], Callable[[int,int],List[Tuple[int]]],
                                                 List[Callable[[int,int],List[Tuple[int]]]]]]=None,
                parameter_symbol:str='θ',
                final_rotation_layer:bool=False,
                name:str=None):
        super().__init__(n_qubit, name=name)
        self._copies = copies
        self.rotation_blocks = rotation_blocks
        self.entanglement_blocks = entanglement_blocks
        self._entangle_strategy = entangle_strategy or 'full' 
        self._parameter_symbol = parameter_symbol
        self._dynamic_param = None
        self._final_rotation_layer = final_rotation_layer
        
    @property
    def copies(self) -> int:
        return self._copies
    
    @property
    def parameter_symbol(self) -> str:
        return self._parameter_symbol
    
    @property
    def entangle_strategy(self):
        return self._entangle_strategy
    
    @property
    def final_rotation_layer(self):
        return self._final_rotation_layer
    
    @property
    def rotation_blocks(self):
        '''A collection of single qubit gate operations or 
        circuit blocks consisting of multiple single qubit operations
        '''
        return self._rotation_blocks
    
    @rotation_blocks.setter
    def rotation_blocks(self, val):
        val = ParameterisedCircuit._parse_blocks(val)
        ParameterisedCircuit._validate_blocks(val, SingleQubitGate, allow_template=False)
        self._rotation_blocks = val
    
    @property
    def entanglement_blocks(self):
        '''A collection of two qubit gate operations or 
        circuit blocks consisting of multiple single/two qubit operations
        '''
        return self._entanglement_blocks
    
    @entanglement_blocks.setter
    def entanglement_blocks(self, val):
        val = ParameterisedCircuit._parse_blocks(val)
        ParameterisedCircuit._validate_blocks(val, TwoQubitGate)
        self._entanglement_blocks = val
        
    @staticmethod
    def _parse_blocks(blocks) -> List:
        if not blocks:
            return []
        blocks = [blocks] if not isinstance(blocks, list) else blocks
        blocks = [QuantumCircuit._parse_gate_operation(block) if
                  isinstance(block, str) else block for block in blocks]
        return blocks
    
    @staticmethod
    def _validate_blocks(blocks, gate_feature=None, allow_template=True) -> None:
        for block in blocks:
            if isinstance(block, TemplateCircuitBlock):
                if not allow_template:
                    raise ValueError('template circuit block is not allowed'
                                     'in rotation layer')
                else:
                    continue
            elif isinstance(block, cirq.Gate):
                if gate_feature and \
                    not isinstance(block, gate_feature):
                    raise ValueError('Gate operation {} should be a subclass of '
                                     '{}'.format(type(block), gate_feature))
            elif callable(block):
                n_param = ParameterisedCircuit._get_parameter_count(block)
                if not n_param:
                    raise ValueError('not a valid gate operation or'
                                     'circuit block: {}'.format(block))
                dummy_params = tuple(sp.symarray('x',n_param))
                ParameterisedCircuit._validate_blocks([block(*dummy_params)], gate_feature)
        
    @staticmethod
    def _get_parameter_count(gate_expr) -> int:
        '''Returns the number of parameters required for a gate operation expression
        a gate operation expression can be a cirq.Gate instance, a string that can be mapped
        to a cirq.Gate instance, a function that maps to cirq.Gate or a TemplateCircuitBlock instance
        '''
        if isinstance(gate_expr, cirq.Gate):
            return 0
        elif isinstance(gate_expr, TemplateCircuitBlock):
            return gate_expr.num_params
        elif callable(gate_expr):
            return len(signature(gate_expr).parameters)
        else:
            raise ValueError('invalid gate expression {} of type {}'.format(gate_expr, type(gate_expr)))
    
    def get_total_parameter_count(self) -> int:
        rotation_param_count = 0
        entanglement_param_count = 0
        for block in self._rotation_blocks:
            rotation_param_count += ParameterisedCircuit._get_parameter_count(block)*self.n_qubit
        for i, block in enumerate(self._entanglement_blocks):
            if isinstance(block, TemplateCircuitBlock):
                connectivity_graphs = self.get_connectivity_graphs(i, block.num_block_qubits)
            else:
                connectivity_graphs = self.get_connectivity_graphs(i, 2)
            entanglement_param_count += ParameterisedCircuit._get_parameter_count(block)*len(connectivity_graphs)
        param_count = (rotation_param_count+entanglement_param_count)*self.copies + \
                        rotation_param_count*self.final_rotation_layer
        return param_count
     
    def reserve_parameters(self) -> None:
        param_count = self.get_total_parameter_count()
        if param_count:
            self._parameter_table.append_array(self.parameter_symbol, param_count)
            self._dynamic_param = iter(self._parameter_table[self.parameter_symbol])
    
    @property
    def dynamic_param(self) -> Optional[Iterator]:
        return self._dynamic_param
    
    def new_param(self) -> sp.Symbol:
        param = next(self.dynamic_param, None)
        if param:
            return param
        else:
            raise ValueError('incorrect number of parameters used in the parameterised circuit')
    
    def new_layer(self, name:str=None):
        layer = super().new_layer(name=name)
        layer._parameter_symbol = self.parameter_symbol
        layer._parameter_table = self._parameter_table
        layer._dynamic_param = self._dynamic_param
        return layer
        
    #https://cirq.readthedocs.io/en/latest/google/best_practices.html
    @staticmethod
    def build_params_for_sweep(self):
        pass
    
    #https://cirq.readthedocs.io/en/latest/google/best_practices.html
    def flatten_sweep(self):
        pass
    
    
    def assemble(self):
        self.clear()
        self.reserve_parameters()
        for _ in range(self.copies):
            # can't just multiply since parameters won't get incremented
            self += self.get_rotation_layer()
            self += self.get_entanglement_layer()
        if self.final_rotation_layer:
            self += self.get_rotation_layer()
                         
    def get_connectivity_graphs(self, block_index:int, num_block_qubits:int, *args, **kwargs):
        strategy = self.entangle_strategy
        if isinstance(strategy, str):
            return connectivity_graph[strategy](self.n_qubit, num_block_qubits)
        elif callable(strategy):
            return strategy(self.n_qubit, num_block_qubits)
        elif isinstance(strategy, list):
            if all(isinstance(strat, str) for strat in strategy):
                return connectivity_graph[strategy[block_index]](self.n_qubit, num_block_qubits)
            elif all(callable(strat) for strat in strategy):
                return strategy[block_index](self.n_qubit, num_block_qubits)
        else:
            raise ValueError('invalid entangle strategy: {}'.format(strategy))
        
    # Devices are generally calibrated to circuits that alternate single-qubit gates with two-qubit gates in each layer. Staying close to this paradigm will often improve performance of circuits.

    def parameterise_gate(self, gate:Union[cirq.Gate, Callable]):
        if isinstance(gate, cirq.Gate):
            return gate
        else:
            n_param = ParameterisedCircuit._get_parameter_count(gate)
            params = tuple(self.new_param() for _ in range(n_param))
            return gate(*params)
    
    def get_rotation_layer(self):
        rotation_layer = self.new_layer('rotation_layer')
        for rotation_block in self.rotation_blocks:
            for qubit in range(self.n_qubit):
                gate = self.parameterise_gate(rotation_block)
                rotation_layer.apply_gate_operation(gate, qubit)
        return rotation_layer

    def get_entanglement_layer(self):
        entanglement_layer = self.new_layer('entanglement_layer')
        for i, entanglement_block in enumerate(self.entanglement_blocks):
            if isinstance(entanglement_block, TemplateCircuitBlock):
                connectivity_graphs = self.get_connectivity_graphs(i, entanglement_block.num_block_qubits)
                for qubits in connectivity_graphs:
                    entanglement_block.build(entanglement_layer, qubits)
            else:
                connectivity_graphs = self.get_connectivity_graphs(i, 2)
                for qubits in connectivity_graphs:
                    gate = self.parameterise_gate(entanglement_block)
                    entanglement_layer.apply_gate_operation(gate, qubits)
    
        return entanglement_layer
        


from functools import reduce
import numpy as np

def self_product(x: np.ndarray) -> float:
    """
    Define a function map from R^n to R.

    Args:
        x: data

    Returns:
        float: the mapped value
    """
    coeff = x[0] if len(x) == 1 else reduce(lambda m, n: m * n, np.pi - x)  
    return coeff


from typing import List, Union, Optional, Callable, Sequence
import numpy as np
from pdb import set_trace

import cirq

from quple import QuantumCircuit, TemplateCircuitBlock
from quple.data_encoding.encoding_circuit import EncodingCircuit

     
                

class PauliExpansion(EncodingCircuit):
    '''PauliExpansion feature map

    '''    
    def __init__(self, feature_dimension: int,
                 depth: int=2, paulis:List[str] = ['Z', 'ZZ'],
                 encoding_map:Optional[Callable[[np.ndarray], float]] = None,
                 name:str=None):
        '''Create PauliExpansion feature map
        Args:
            feature_dimension: dimension of data to be encoded (=number of qubits in the circuit)
            depth: the number of repetition of the encoding circuit
            paulis: pauli operations to be performed on each entangling block
            encoding_map: data mapping function from R^(feature_dimension) to R
            name: name of circuit
        '''
        super().__init__(feature_dimension, depth, encoding_map, name=name)
        self.paulis = paulis
        
    @staticmethod
    def _validate_paulis(paulis:List[str]):
        for pauli_str in paulis:
            for pauli in pauli_str:
                if pauli not in ['Z','X','Y','I']:
                    raise ValueError('Invalid Pauli operation: {}'.format(pauli))
        
    @property
    def paulis(self):
        return self._paulis
    
    @paulis.setter
    def paulis(self, value):
        PauliExpansion._validate_paulis(value)
        self._paulis = value
    
    @staticmethod
    def change_basis(circuit:QuantumCircuit, qubits:Sequence[int],
                     pauli_string:Sequence[str], inverse=False) -> None:
        for i, pauli in enumerate(pauli_string):
            if pauli == 'X':
                circuit.H(qubits[i])
            elif pauli == 'Y':
                circuit.RX(-np.pi / 2 if inverse else np.pi / 2, qubits[i])
    
    
                
        
        
    def _build_primary_layer(self):
        # build the Hadamard layer
        # this maps the |0>^n state into a superposition of 2^n states of equal weight
        # i.e. to the computational basis
        self.H(range(self._n_qubit))
                     
        # build the entanglement layer
        for pauli_string in self.paulis:
            pauli_string = pauli_string[::-1]
            n_pauli = len(pauli_string)
            indices = [i for i, pauli in enumerate(pauli_string) if pauli != 'I']
            if not indices:
                continue
                
            entangler_map = self.get_entangler_map(n_pauli, method='full')
            for qubits in entangler_map:
                
                PauliExpansion.change_basis(self, qubits, pauli_string)
                qubits_to_entangle = tuple(qubits[i] for i in indices)
                encoded_value = self.encode_parameters(qubits) 
                
                self.entangle(qubits_to_entangle)    
                self.RZ(2.0*encoded_value, qubits[-1])
                self.entangle(qubits_to_entangle, inverse=True)
                
                PauliExpansion.change_basis(self, qubits, pauli_string, inverse=True)
        
def self_product(x: np.ndarray) -> float:
    """
    Define a function map from R^n to R.

    Args:
        x: data

    Returns:
        float: the mapped value
    """
    coeff = x[0] if len(x) == 1 else reduce(lambda m, n: m * n, np.pi - x)  
    return coeff



from typing import List, Union, Optional, Callable, Sequence, Tuple

from quple import ParameterisedCircuit

class ExcitationPreserving(ParameterisedCircuit):
    '''Excitation preserving trial wavefunction

    '''    
    def __init__(self, n_qubit: int, copies: int=2, 
                 entanglement_gate:str='RISWAP',
                 entangle_strategy:Optional[Union[str,List[str], Callable[[int,int],List[Tuple[int]]],
                                                 List[Callable[[int,int],List[Tuple[int]]]]]]=None,
                 parameter_symbol:str='θ', name:str=None):
        
        allowed_blocks = ['RISWAP', 'FSim']
        if entanglement_gate not in allowed_blocks:
            raise ValueError('Unsupported gate operation {}, choose one of '
                             '{}'.format(entanglement_gate, allowed_blocks))
 
        super().__init__(n_qubit=n_qubit, copies=copies,
                         rotation_blocks='RZ',
                         entanglement_blocks=entanglement_gate,
                         entangle_strategy=entangle_strategy,
                         parameter_symbol=parameter_symbol,
                         final_rotation_layer=True, name=name)


from typing import Sequence
import numpy as np

from quple.circuit.quantum_circuit import QuantumCircuit


def change_rz_basis(circuit:QuantumCircuit, qubits:Sequence[int], pauli_string:str, inverse=False) -> None:
    '''re-express a parametrised rotation along X and Y axis in terms of 
    parametrised rotation along Z axise
    Example:
        Rx(pi-x[0]) == H\otimes Rz(pi-x[0]) \otimes H
    '''
    for i, pauli in enumerate(pauli_string):
        if pauli == 'X':
            circuit.H(qubits[i])
        elif pauli == 'Y':
            circuit.Rx(-np.pi / 2 if inverse else np.pi / 2, i)

            
            
            
def google_device_layouts(name):
    if name == 'Syncamore':
        return cirq.google.Syncamore
    elif name == 'Sycamore23':
        return cirq.google.Sycamore23
    elif name == 'Bristlecone':
        return cirq.google.Bristlecone
    elif name == 'Foxtail':
        return cirq.google.Foxtail


from typing import Union

def bounded_below(value:Union[float, int], bound:Union[float, int], label:str='') -> None:
    if value < bound:
        raise ValueError('Value out of bound. ({} < {})'.format(label, bound))

def bounded_above(value:Union[float, int], bound:Union[float, int], label:str='') -> None:
    if value > bound:
        raise ValueError('Value out of bound. ({} > {})'.format(label, bound))
        
def bounded_between(value:Union[float, int], lower:Union[float, int], upper:Union[float, int], label:str='') -> None:
    if not ( lower < value < upper):
        raise ValueError('Value out of bound. ( {} < {} < {})'.format(lower, label, upper))
        


from typing import Callable, Optional
import numpy as np

from quple.data_encoding.pauli_expansion import PauliExpansion

class PauliZExpansion(PauliExpansion):
    def __init__(self, feature_dimension: int,
                 depth: int=2, z_order:int = 2,
                 encoding_map:Optional[Callable[[np.ndarray], float]] = None,
                 name:str=None):
        '''Create PauliZExpansion feature map
        Args:
            feature_dimension: dimension of data to be encoded (=number of qubits in the circuit)
            depth: the number of repetition of the encoding circuit
            z_order: order of pauli z operations to be performed on each entangling block
            encoding_map: data mapping function from R^(feature_dimension) to R
            name: name of circuit
        '''        
        paulis = []
        for i in range(1, z_order + 1):
            paulis.append('Z' * i)
        super().__init__(feature_dimension, depth, paulis=paulis,
                         encoding_map=encoding_map, name=name)


from typing import (Any, Callable, cast, Dict, FrozenSet, Iterable, Iterator,
                    List, Optional, overload, Sequence, Set, Tuple, Type,
                    TYPE_CHECKING, TypeVar, Union)    

import re
import sympy as sp
import numpy as np

class ParameterTable():
    def __init__(self, params: Union[str, List[str], List[sp.Symbol]]=None):
        self._params = {}
        if params:
            self.append(params)
    
    @property
    def params(self) -> Dict:
        return self._params
    
    def count_params(self) -> int:
        count = 0
        for name in self._params:
            if isinstance(self._params[name], np.ndarray):
                count += self._params[name].shape[0]
            else:
                count += 1
        return count
    
    def count_symbols(self) -> int:
        return len(self._params)
    
    def pop(self, symbol:str):
        self._params.pop(symbol, None)
    
    def __getitem__(self, key):
        return self._params[key]
    
    def __len__(self):
        return len(self._params)
    
    
    def append_array(self, symbol:str, size:int):
        self._params[symbol] = sp.symarray(symbol, size)
        
    def keys(self):
        return self._params.keys()
    
    def values(self):
        return self._params.values()
    
    def append(self, params: Union[str, List[str], sp.Symbol, List[sp.Symbol]] ):
        if isinstance(params, list):
            if all(isinstance(param, str) for param in params):
                self._params.update(UnresolvedParameters._create_param_key_values(params))
            elif all(isinstance(param, sp.Symbol) for param in params):
                self._params.update({param.name: param for param in params})
            else:
                raise ValueError('inconsistent list of parameters passed: {}'.format(params))
        elif isinstance(params, str):
            self.append([params])
        elif isinstance(params, sp.Symbol):
            self._params.update({params.name: params})
        else:
            raise TypeError('unsupported parameter type : {}'.format(type(params)))
    
    @staticmethod
    def _create_param_key_values(params:List[str]) -> Dict:
        return {param: sp.Symbol(param) for param in params}
    
    
        


from typing import List, Union, Optional, Callable, Sequence, Tuple

from quple import ParameterisedCircuit

class EfficientSU2(ParameterisedCircuit):
    '''Efficient SU2 trial wavefunction

    '''    
    def __init__(self, n_qubit: int, copies: int=2, 
                 su2_gates:Optional[Union[str, 'cirq.Gate', Callable, 'TemplateCircuitBlock',
                                               List[str],List['cirq.Gate'],List[Callable],
                                               List['TemplateCircuitBlock']]]=None,
                 entangle_strategy:Optional[Union[str,List[str], Callable[[int,int],List[Tuple[int]]],
                                                 List[Callable[[int,int],List[Tuple[int]]]]]]=None,
                 parameter_symbol:str='θ', name:str=None):
        
        su2_gates = su2_gates or ['RY','RZ']
        super().__init__(n_qubit=n_qubit, copies=copies,
                         rotation_blocks=su2_gates,
                         entanglement_blocks='CX',
                         entangle_strategy=entangle_strategy,
                         parameter_symbol=parameter_symbol,
                         final_rotation_layer=True, name=name)


from abc import ABC, abstractmethod
from typing import List, Sequence


class TemplateCircuitBlock(ABC):
    
    @abstractmethod
    def build(self, circuit:'ParameterisedCircuit', qubits:Sequence[int]):
        '''Builds the circuit block involving the specified qubits
        '''
        raise NotImplementedError
    
    @property
    @abstractmethod
    def num_params(self) -> int:
        """Returns the number of parameters required by the circuit block.
        """
        
    @property
    @abstractmethod
    def num_block_qubits(self) -> int:
        """Returns the number of qubits in the circuit block
        """